"""Command line interface for kg-chat."""

import logging
import shutil
from pathlib import Path
from pprint import pprint
from typing import Union

import click

from kg_chat import __version__
from kg_chat.app import create_app
from kg_chat.constants import OPEN_AI_MODEL, VECTOR_DB_PATH, VECTOR_STORE
from kg_chat.main import KnowledgeGraphChat
from kg_chat.utils import (
    get_anthropic_models,
    get_database_impl,
    get_lbl_cborg_models,
    get_llm_config,
    get_ollama_models,
    get_openai_models,
)

__all__ = [
    "main",
]

ALL_AVAILABLE_PROVIDERS = ["openai", "ollama", "anthropic", "cborg"]
ALL_AVAILABLE_MODELS = get_openai_models() + get_ollama_models() + get_anthropic_models() + get_lbl_cborg_models()
ALL_AVAILABLE_DB = ["neo4j", "duckdb"]

logger = logging.getLogger(__name__)
database_options = click.option(
    "--database",
    "-d",
    type=click.Choice(ALL_AVAILABLE_DB, case_sensitive=False),
    help="Database to use.",
    default="duckdb",
)
data_dir_option = click.option(
    "--data-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Directory containing the data.",
    required=True,
)
docs_option = click.option(
    "--docs",
    type=click.Path(exists=True, file_okay=True, dir_okay=True),
    help="Path to a document or directory of only documents.",
    required=False,
    default=None,
)
llm_provider_option = click.option(
    "--llm-provider",
    type=click.Choice(ALL_AVAILABLE_PROVIDERS, case_sensitive=False),
    help="Language model provider to use.",
    required=False,
)
llm_option = click.option(
    "--llm",
    type=click.Choice(ALL_AVAILABLE_MODELS, case_sensitive=False),
    help="Language model to use.",
    required=False,
)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
@click.version_option(__version__)
def main(verbose: int, quiet: bool):
    """
    CLI for kg-chat.

    :param verbose: Verbosity while running.
    :param quiet: Boolean to be quiet or verbose.
    """
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)


@main.command()
def list_models():
    """List the available language models."""
    openai_models = "\n  ".join(get_openai_models())
    anthropic_models = "\n  ".join(get_anthropic_models())
    ollama_models = "\n  ".join(get_ollama_models())
    lbl_cborg_models = "\n  ".join(get_lbl_cborg_models())

    click.echo(f"OpenAI models:\n  {openai_models}")
    click.echo(f"Anthropic models:\n  {anthropic_models}")
    click.echo(f"Ollama models:\n  {ollama_models}")
    click.echo(f"LBL-CBORG models:\n  {lbl_cborg_models}")


@main.command("import")
@database_options
@data_dir_option
@docs_option
@llm_provider_option
def import_kg(database: str = "duckdb", data_dir: str = None, docs: str = None, llm_provider: str = "openai"):
    """Run the kg-chat's import command."""
    if docs and VECTOR_DB_PATH.exists():
        for item in VECTOR_STORE.iterdir():
            if item.is_dir():
                shutil.rmtree(item)
            else:
                item.unlink()

    config = get_llm_config(llm_provider)
    impl = get_database_impl(database, data_dir=data_dir, doc_dir_or_file=docs, llm_config=config)
    impl.load_kg()


@main.command()
@data_dir_option
@database_options
@llm_provider_option
@llm_option
def test_query(data_dir: Union[str, Path], llm_provider: str, llm: str, database: str = "duckdb"):
    """Run the kg-chat's test-query command."""
    if llm_provider is None and llm is None:
        llm = OPEN_AI_MODEL
    config = get_llm_config(llm_provider, llm)
    impl = get_database_impl(database, data_dir=data_dir, llm_config=config)

    query = "MATCH (n) RETURN n LIMIT 10" if database == "neo4j" else "SELECT * FROM nodes LIMIT 10"
    result = impl.execute_query(query)
    for record in result:
        print(record)


@main.command()
@data_dir_option
@database_options
@llm_provider_option
@llm_option
def show_schema(data_dir: Union[str, Path], llm_provider: str, llm: str, database: str = "duckdb"):
    """Run the kg-chat's show-schema command."""
    config = get_llm_config(llm_provider, llm)
    impl = get_database_impl(database, data_dir=data_dir, llm_config=config)
    impl.show_schema()


@main.command()
@database_options
@click.argument("query", type=str, required=True)
@data_dir_option
@llm_provider_option
@llm_option
def qna(query: str, data_dir: Union[str, Path], llm_provider: str, llm: str, database: str = "duckdb"):
    """Run the kg-chat's qna command."""
    config = get_llm_config(llm_provider, llm)
    impl = get_database_impl(database, data_dir=data_dir, llm_config=config)
    response = impl.get_human_response(query)
    pprint(response)


@main.command("chat")
@data_dir_option
@database_options
@llm_provider_option
@llm_option
def run_chat(data_dir: Union[str, Path], llm_provider: str, llm: str, database: str = "duckdb"):
    """Run the kg-chat's chat command."""
    config = get_llm_config(llm_provider, llm)
    impl = get_database_impl(database, data_dir=data_dir, llm_config=config)
    kgc = KnowledgeGraphChat(impl)
    kgc.chat()


@main.command("app")
@click.option("--debug", is_flag=True, help="Run the app in debug mode.")
@data_dir_option
@database_options
@llm_provider_option
@llm_option
def run_app(
    data_dir: Union[str, Path],
    llm_provider: str,
    llm: str,
    database: str = "duckdb",
    debug: bool = False,
):
    """Run the kg-chat's app command."""
    config = get_llm_config(llm_provider, llm)
    impl = get_database_impl(database, data_dir=data_dir, llm_config=config)
    kgc = KnowledgeGraphChat(impl)
    app = create_app(kgc)
    # use_reloader=False to avoid running the app twice in debug mode
    app.run(debug=debug, use_reloader=False)


if __name__ == "__main__":
    main()



"""Implementation of the DatabaseInterface for Neo4j."""

import csv
import logging
import time
from pathlib import Path
from pprint import pprint
from typing import Any, Optional, Sequence, Union

from langchain.agents import Tool
from langchain.agents.agent import AgentExecutor
from langchain.agents.agent_types import AgentType
from langchain.agents.loading import AGENT_TO_CLASS, load_agent
from langchain.tools.retriever import create_retriever_tool
from langchain_community.chains.graph_qa.cypher import GraphCypherQAChain
from langchain_community.graphs import Neo4jGraph
from langchain_core.callbacks import BaseCallbackManager
from langchain_core.language_models import BaseLanguageModel
from langchain_core.tools import BaseTool
from langchain_ollama import ChatOllama
from neo4j import GraphDatabase

from kg_chat.config.llm_config import LLMConfig
from kg_chat.constants import DATALOAD_BATCH_SIZE, NEO4J_PASSWORD, NEO4J_URI, NEO4J_USERNAME, VECTOR_DB_PATH
from kg_chat.interface.database_interface import DatabaseInterface
from kg_chat.utils import (
    create_vectorstore,
    get_cypher_agent_prompt_template,
    get_exisiting_vectorstore,
    llm_factory,
    structure_query,
)

logger = logging.getLogger(__name__)


class Neo4jImplementation(DatabaseInterface):
    """Implementation of the DatabaseInterface for Neo4j."""

    def __init__(
        self,
        data_dir: Union[str, Path],
        doc_dir_or_file: Union[str, Path] = None,
        uri: str = NEO4J_URI,
        username: str = NEO4J_USERNAME,
        password: str = NEO4J_PASSWORD,
        llm_config: LLMConfig = None,
    ):
        """Initialize the Neo4j database and the Langchain components."""
        if not data_dir:
            raise ValueError("Data directory is required. This typically contains the KGX tsv files.")
        self.driver = GraphDatabase.driver(uri, auth=(username, password))
        self.graph = Neo4jGraph(url=uri, username=username, password=password)
        self.llm = llm_factory(llm_config)
        self.tools = []
        self.data_dir = Path(data_dir)
        if VECTOR_DB_PATH.exists() and get_exisiting_vectorstore():
            vectorstore = get_exisiting_vectorstore()
            retriever = vectorstore.as_retriever(search_kwargs={"k": 1})
            rag_tool = create_retriever_tool(retriever, "VectorStoreRetriever", "Vector Store Retriever")

            self.tools.append(rag_tool)
        elif doc_dir_or_file:
            vectorstore = create_vectorstore(doc_dir_or_file=doc_dir_or_file)
            retriever = vectorstore.as_retriever(search_kwargs={"k": 1})
            rag_tool = create_retriever_tool(retriever, "VectorStoreRetriever", "Vector Store Retriever")
            self.tools.append(rag_tool)
        else:
            logger.info("No vectorstore found or documents provided. Skipping RAG tool creation.")

        graph_cypher_qa_chain = GraphCypherQAChain.from_llm(
            graph=self.graph,
            llm=self.llm,
            verbose=True,
            use_function_response=True,
            # function_response_system="Respond as a Data Scientist!",
            validate_cypher=True,
        )
        graph_cypher_qa_chain_tool = Tool(
            name="GraphCypherQAChain",
            description="Graph Cypher QA Chain",
            func=graph_cypher_qa_chain.invoke,
        )

        self.tools.append(graph_cypher_qa_chain_tool)
        self.tool_names = [tool.name for tool in self.tools]
        self.safe_mode = True
        self.agent_executor = self.initialize_agent(
            llm=self.llm,
            tools=self.tools,
            prompt=get_cypher_agent_prompt_template(),
            handle_parsing_errors=True,
            verbose=True,
        )

        # self.agent_executor = AgentExecutor(
        #     agent=self.agent, tools=self.tools, handle_parsing_errors=True, verbose=True
        # )
        logger.info("Agent executor created successfully.")

    # ! Adoped from langchain.agents.initialize.initialize_agent since it'll be deprecated 1.0 onwards.
    # ! This works and `create_react_agent` which is the alternative doesn't work.
    @staticmethod
    def initialize_agent(
        tools: Sequence[BaseTool],
        llm: BaseLanguageModel,
        agent: Optional[AgentType] = None,
        callback_manager: Optional[BaseCallbackManager] = None,
        agent_path: Optional[str] = None,
        agent_kwargs: Optional[dict] = None,
        *,
        tags: Optional[Sequence[str]] = None,
        **kwargs: Any,
    ) -> AgentExecutor:
        """
        Load an agent executor given tools and LLM.

        :param tools: List of tools this agent has access to.
        :param llm: Language model to use as the agent.
        :param agent: Agent type to use. If None and agent_path is also None, will default
                    to AgentType.ZERO_SHOT_REACT_DESCRIPTION. Defaults to None.
        :param callback_manager: CallbackManager to use. Global callback manager is used if
                                not provided. Defaults to None.
        :param agent_path: Path to serialized agent to use. If None and agent is also None,
                        will default to AgentType.ZERO_SHOT_REACT_DESCRIPTION. Defaults to None.
        :param agent_kwargs: Additional keyword arguments to pass to the underlying agent.
                            Defaults to None.
        :param tags: Tags to apply to the traced runs. Defaults to None.
        :param kwargs: Additional keyword arguments passed to the agent executor.

        :returns: An agent executor.

        :raises ValueError: If both `agent` and `agent_path` are specified.
        :raises ValueError: If `agent` is not a valid agent type.
        :raises ValueError: If both `agent` and `agent_path` are None.
        """
        tags_ = list(tags) if tags else []
        if agent is None and agent_path is None:
            agent = AgentType.ZERO_SHOT_REACT_DESCRIPTION
        if agent is not None and agent_path is not None:
            raise ValueError("Both `agent` and `agent_path` are specified, " "but at most only one should be.")
        if agent is not None:
            if agent not in AGENT_TO_CLASS:
                raise ValueError(f"Got unknown agent type: {agent}. " f"Valid types are: {AGENT_TO_CLASS.keys()}.")
            tags_.append(agent.value if isinstance(agent, AgentType) else agent)
            agent_cls = AGENT_TO_CLASS[agent]
            agent_kwargs = agent_kwargs or {}
            agent_obj = agent_cls.from_llm_and_tools(llm, tools, callback_manager=callback_manager, **agent_kwargs)
        elif agent_path is not None:
            agent_obj = load_agent(agent_path, llm=llm, tools=tools, callback_manager=callback_manager)
            try:
                # TODO: Add tags from the serialized object directly.
                tags_.append(agent_obj._agent_type)
            except NotImplementedError:
                pass
        else:
            raise ValueError("Somehow both `agent` and `agent_path` are None, " "this should never happen.")

        return AgentExecutor.from_agent_and_tools(
            agent=agent_obj,
            tools=tools,
            callback_manager=callback_manager,
            tags=tags_,
            **kwargs,
        )

    def toggle_safe_mode(self, enabled: bool):
        """Toggle safe mode on or off."""
        self.safe_mode = enabled

    def is_safe_command(self, command: str) -> bool:
        """Check if a command is safe to execute."""
        unsafe_keywords = ["CREATE", "DELETE", "REMOVE", "SET", "MERGE", "DROP"]
        return not any(keyword in command.upper() for keyword in unsafe_keywords)

    def execute_unsafe_operation(self, operation: callable, *args, **kwargs):
        """Execute an unsafe operation."""
        original_safe_mode = self.safe_mode
        self.safe_mode = False
        try:
            result = operation(*args, **kwargs)
        finally:
            self.safe_mode = original_safe_mode
        return result

    def execute_query(self, query: str):
        """Execute a Cypher query against the Neo4j database."""
        if self.safe_mode and not self.is_safe_command(query):
            raise ValueError(f"Unsafe command detected: {query}")
        with self.driver.session() as session:
            result = session.read_transaction(lambda tx: list(tx.run(query)))
        return result

    def execute_query_using_langchain(self, query: str):
        """Execute a Cypher query against the Neo4j database using Langchain."""
        if self.safe_mode and not self.is_safe_command(query):
            raise ValueError(f"Unsafe command detected: {query}")
        result = self.graph.query(query)
        return result

    def clear_database(self):
        """Clear the Neo4j database."""

        def _clear_database():
            with self.driver.session() as session:
                session.write_transaction(self._clear_database_tx)

        return self.execute_unsafe_operation(_clear_database)

    @staticmethod
    def _clear_database_tx(tx):
        """Clear the Neo4j database using APOC."""
        tx.run(
            """
        CALL apoc.periodic.iterate(
            'MATCH (n) RETURN n',
            'DETACH DELETE n',
            {batchSize:1000, parallel:true}
        )
        """
        )

    def ensure_index(self):
        """Ensure that the index on :Node(id) exists."""

        def _ensure_index():
            with self.driver.session() as session:
                session.write_transaction(self._ensure_index_tx)

        return self.execute_unsafe_operation(_ensure_index)

    @staticmethod
    def _ensure_index_tx(tx):
        """Index existence check."""
        tx.run(
            """
            CALL apoc.schema.assert(
                {Node: [['id'], ['category'], ['label']]},
                {RELATIONSHIP: [['subject'], ['predicate'], ['object']]}
            )
            """
        )
        print("Indexes ensured using APOC.")

    def get_human_response(self, prompt: str):
        """Get a human response from the Neo4j database."""
        # human_response = self.chain.invoke({"query": prompt})
        human_response = self.agent_executor.invoke(
            {
                "input": prompt,
                "tools": self.tools,
                "tool_names": self.tool_names,
            }
        )

        pprint(human_response["output"])
        return human_response["output"]

    def get_structured_response(self, prompt: str):
        """Get a structured response from the Neo4j database."""
        if isinstance(self.llm, ChatOllama):
            if "show me" in prompt.lower():
                self.llm.format = "json"
        # response = self.chain.invoke({"query": structure_query(prompt)})
        response = self.agent_executor.invoke(
            {
                "input": structure_query(prompt),
                "tools": self.tools,
                "tool_names": self.tool_names,
            }
        )
        return response["output"]

    def create_edges(self, edges):
        """Create relationships between nodes."""

        def _create_edges():
            with self.driver.session() as session:
                session.write_transaction(self._create_edges_tx, edges)

        return self.execute_unsafe_operation(_create_edges)

    @staticmethod
    def _create_edges_tx(tx, edges):
        tx.run(
            """
            UNWIND $edges AS edge
            MATCH (a:Node {id: edge.subject})
            MATCH (b:Node {id: edge.object})
            CREATE (a)-[r:RELATIONSHIP {type: edge.predicate}]->(b)
        """,
            edges=edges,
        )

    def create_nodes(self, nodes):
        """Create nodes in the Neo4j database."""

        def _create_nodes():
            with self.driver.session() as session:
                session.write_transaction(self._create_nodes_tx, nodes)

        return self.execute_unsafe_operation(_create_nodes)

    @staticmethod
    def _create_nodes_tx(tx, nodes):
        tx.run(
            """
            UNWIND $nodes AS node
            CREATE (n:Node {id: node.id, category: node.category, label: node.label})
        """,
            nodes=nodes,
        )

    def show_schema(self):
        """Show the schema of the Neo4j database."""
        with self.driver.session() as session:
            result = session.read_transaction(lambda tx: list(tx.run("CALL db.schema.visualization()")))
            pprint(result)

    def load_kg(self, block_size: int = DATALOAD_BATCH_SIZE):
        """Load the Knowledge Graph into the Neo4j database."""
        nodes_filepath = self.data_dir / "nodes.tsv"
        edges_filepath = self.data_dir / "edges.tsv"

        def _load_kg():
            # Clear the existing database
            print("Clearing the existing database...")
            self.clear_database()
            print("Database cleared.")

            # Ensure indexes are created
            print("Ensuring indexes...")
            self.ensure_index()
            print("Indexes ensured.")

            # Import nodes in batches
            print("Starting to import nodes...")
            start_time = time.time()
            nodes_batch = []
            columns_of_interest = ["id", "category", "name", "description"]

            with open(nodes_filepath, "r") as nodes_file:
                reader = csv.DictReader(nodes_file, delimiter="\t")
                node_batch_loaded = 0

                # Determine which label column to use
                label_column = "name" if "name" in reader.fieldnames else "description"

                for row in reader:
                    node_id = row[columns_of_interest[0]]
                    node_category = row[columns_of_interest[1]]
                    node_label = row[label_column]
                    nodes_batch.append({"id": node_id, "category": node_category, "label": node_label})
                    node_batch_loaded += 1

                    if len(nodes_batch) >= block_size:
                        self.create_nodes(nodes_batch)
                        nodes_batch = []

                if nodes_batch:
                    self.create_nodes(nodes_batch)

            elapsed_time_seconds = time.time() - start_time

            if elapsed_time_seconds >= 3600:
                elapsed_time_hours = elapsed_time_seconds / 3600
                print(f"Nodes import completed: {node_batch_loaded} nodes in {elapsed_time_hours:.2f} hours.")
            else:
                elapsed_time_minutes = elapsed_time_seconds / 60
                print(f"Nodes import completed: {node_batch_loaded} nodes in {elapsed_time_minutes:.2f} minutes.")

            # Import edges in batches
            print("Starting to import edges...")
            start_time = time.time()
            edges_batch = []
            edge_column_of_interest = ["subject", "predicate", "object"]

            with open(edges_filepath, "r") as edges_file:
                reader = csv.DictReader(edges_file, delimiter="\t")
                edge_batch_loaded = 0

                for row in reader:
                    subject = row[edge_column_of_interest[0]]
                    predicate = row[edge_column_of_interest[1]]
                    object = row[edge_column_of_interest[2]]
                    edges_batch.append({"subject": subject, "predicate": predicate, "object": object})
                    edge_batch_loaded += 1

                    if len(edges_batch) >= block_size / 2:
                        self.create_edges(edges_batch)
                        edges_batch = []

                if edges_batch:
                    self.create_edges(edges_batch)

            elapsed_time_seconds = time.time() - start_time
            if elapsed_time_seconds >= 3600:
                elapsed_time_hours = elapsed_time_seconds / 3600
                print(f"Edges import completed: {edge_batch_loaded} edges in {elapsed_time_hours:.2f} hours.")
            else:
                elapsed_time_minutes = elapsed_time_seconds / 60
                print(f"Edges import completed: {edge_batch_loaded} edges in {elapsed_time_minutes:.2f} minutes.")

            print("Import process finished.")

        return self.execute_unsafe_operation(_load_kg)

    def __del__(self):
        """Ensure the driver is closed when the object is destroyed."""
        if hasattr(self, "driver"):
            self.driver.close()



"""Utility functions for the KG chatbot."""

import logging
import os
import random
import webbrowser
from pathlib import Path
from typing import Union

from langchain_chroma import Chroma
from langchain_core.documents import Document
from langchain_core.prompts.prompt import PromptTemplate
from langchain_openai import OpenAIEmbeddings
from langchain_text_splitters import RecursiveCharacterTextSplitter
from openai import OpenAI
from pyvis.network import Network

from kg_chat.config.llm_config import AnthropicConfig, CBORGConfig, LLMConfig, OllamaConfig, OpenAIConfig
from kg_chat.constants import (
    ANTHROPIC_KEY,
    ANTHROPIC_MODEL,
    CBORG_API_KEY,
    CBORG_MODEL,
    OLLAMA_MODEL,
    OPEN_AI_MODEL,
    OPENAI_KEY,
    VECTOR_DB_PATH,
    VECTOR_STORE,
)

PREFIX_COLOR_MAP = {}
logger = logging.getLogger(__name__)


def extract_nodes_edges(structured_result):
    """Extract nodes and edges from the structured result."""
    nodes = structured_result.get("nodes", [])
    edges = structured_result.get("edges", [])

    return nodes, edges


# * LLM related utilities.**************************************************************************
def structure_query(query: str) -> str:
    """Structure the query to request structured results."""
    if "show me" in query.lower():
        # Modify the query to request structured results
        structured_query = f"""
        {query}
        Please provide the result in JSON format with ALL nodes and ALL edges. Return JSON ONLY.
        Property name MUST be enclosed in double quotes.
        IMPORTANT: I want nothing but pure JSON (verbose=False). Remove all unnecessary quotes and newlines.
        Example: {{
            "nodes": [
                {{"label": "A", "id": "1", "category": "biolink:Gene"}},
                {{"label": "B", "id": "2", "category": "biolink:ChemicalSubstance"}},
                {{"label": "C", "id": "3", "category": "biolink:Disease"}}
            ],
            "edges": [
                {{"subject": {{"label": "A", "id": "1"}},
                "object": {{"label": "B", "id": "2"}},
                "predicate": "biolink:related_to"}}
            ]
        }}
        """
    else:
        structured_query = query
    return structured_query


def get_sql_agent_prompt_template():
    """Get the agent prompt."""
    template = """
                You are an amazing Data Scientist that queries DuckDB databases by executing SQL queries
                generated by yourself and return the results from those queries.
                You have access to additional tools like SQLDatabaseToolkit and QuerySQLDataBaseTool
                in order to answer the following questions as best you can. DO NOT fabricate responses.
                Any response should be strictly based on the information provided by the database via queries.
                Always answer in the same language as the user question. The database is a DuckDB database.
                The data is simply two tables : "nodes" and "edges" which forma a knowledge graph. Any question
                asked should be answered based on this knowledge graph. You may or may not have documents
                provided for RAG. If you have documents, you can use them to answer questions.

                You have access to the following tools:

                {tools}

                To use a tool, please use the following format:

                '''
                "Thought": Do I need to use a tool? Yes.
                "Action 1": the action to take, should be one of [{tool_names}].
                "Action Input 1": the input to the action.
                "Observation": the result of the action.
                ... (this Thought/Action/Action Input/Observation can repeat 5 times)
                '''

                When the query returns the result use the format:
                '''
                "Thought": Do I need to use a tool? No
                "Final Answer": [your response here]
                '''

                Begin!

                Question: {input}
                Thought:{agent_scratchpad}
    """

    return PromptTemplate(
        input_variables=["input", "agent_scratchpad", "tools", "tool_names", "intermediate_steps"], template=template
    )


def get_cypher_agent_prompt_template():
    """Get the agent prompt for querying a Neo4J database."""
    template = """
                You are an amazing Data Scientist that queries Neo4J databases by executing Cypher queries
                generated by yourself and return the results from those queries.
                You have access to additional tools like CypherQueryToolkit and QueryCypherDatabaseTool
                in order to answer the following questions as best you can. DO NOT fabricate responses.
                Any response should be strictly based on the information provided by the database via queries.
                Always answer in the same language as the user question. The database is a Neo4J database.
                The data is simply two types of entities: "nodes" and "relationships" which form a knowledge graph.
                Any question asked should be answered based on this knowledge graph. You may or may not have documents
                provided for RAG. If you have documents, you can use them to answer questions.

                You have access to the following tools:

                {tools}

                To use a tool, please use the following format:

                '''
                "Thought": Do I need to use a tool? Yes.
                "Action 1": the action to take, should be one of [{tool_names}].
                "Action Input 1": the input to the action.
                "Observation": the result of the action.
                ... (this Thought/Action/Action Input/Observation can repeat 5 times)
                '''

                When the query returns the result use the format:
                '''
                "Thought": Do I need to use a tool? No
                "Final Answer": [your response here]
                '''

                Begin!

                Question: {input}
                Thought:{agent_scratchpad}
    """

    return PromptTemplate(
        input_variables=["input", "agent_scratchpad", "tools", "tool_names", "intermediate_steps"], template=template
    )


def get_llm_config(llm_provider: str, llm_model: str = None):
    """Get the LLM configuration based on the selected LLM."""

    def validate_and_get_model(provider, default_model, get_models_func):
        """Validate the model and get the model."""
        if llm_model is None:
            return default_model
        list_of_models = get_models_func()
        if llm_model not in list_of_models:
            raise ValueError(f"Model {llm_model} not supported. Please choose from {list_of_models}")
        return llm_model

    if llm_provider == "openai":
        from kg_chat.config.llm_config import OpenAIConfig

        llm_model = validate_and_get_model("openai", OPEN_AI_MODEL, get_openai_models)
        return OpenAIConfig(model=llm_model, api_key=OPENAI_KEY)

    elif llm_provider == "ollama":
        from kg_chat.config.llm_config import OllamaConfig

        llm_model = validate_and_get_model("ollama", OLLAMA_MODEL, get_ollama_models)
        return OllamaConfig(model=llm_model)

    elif llm_provider == "anthropic":
        from kg_chat.config.llm_config import AnthropicConfig

        llm_model = validate_and_get_model("anthropic", ANTHROPIC_MODEL, get_anthropic_models)
        return AnthropicConfig(model=llm_model, api_key=ANTHROPIC_KEY)

    elif llm_provider == "cborg":
        from kg_chat.config.llm_config import CBORGConfig

        llm_model = validate_and_get_model("cborg", CBORG_MODEL, get_lbl_cborg_models)
        return CBORGConfig(model=llm_model, api_key=CBORG_API_KEY)

    else:
        all_models = {
            "openai": get_openai_models(),
            "ollama": get_ollama_models(),
            "anthropic": get_anthropic_models(),
            "cborg": get_lbl_cborg_models(),
        }
        if llm_model is None:
            llm_model = OPEN_AI_MODEL
            from kg_chat.config.llm_config import OpenAIConfig

            return OpenAIConfig(model=llm_model, api_key=OPENAI_KEY)

        for provider, models in all_models.items():
            if llm_model in models:
                return get_llm_config(provider, llm_model)

        raise ValueError(f"Model {llm_model} not supported.")


def get_database_impl(database: str, data_dir: str, llm_config, doc_dir_or_file: str = None):
    """Get the database implementation based on the selected database."""
    if database == "neo4j":
        from kg_chat.implementations.neo4j_implementation import Neo4jImplementation

        return Neo4jImplementation(data_dir=data_dir, doc_dir_or_file=doc_dir_or_file, llm_config=llm_config)
    elif database == "duckdb":
        from kg_chat.implementations.duckdb_implementation import DuckDBImplementation

        return DuckDBImplementation(data_dir=data_dir, doc_dir_or_file=doc_dir_or_file, llm_config=llm_config)
    else:
        raise ValueError(f"Database {database} not supported.")


def llm_factory(config: LLMConfig):
    """Create an LLM instance based on the configuration."""
    if isinstance(config, OpenAIConfig):
        from langchain_openai import ChatOpenAI

        return ChatOpenAI(model=config.model, temperature=config.temperature, api_key=config.api_key)
    elif isinstance(config, OllamaConfig):
        from langchain_ollama import ChatOllama

        return ChatOllama(model=config.model, temperature=config.temperature, api_key=config.api_key)
    elif isinstance(config, AnthropicConfig):
        from langchain_anthropic import ChatAnthropic

        return ChatAnthropic(
            model=config.model, temperature=config.temperature, api_key=config.api_key, max_tokens_to_sample=4096
        )
    elif isinstance(config, CBORGConfig):
        from langchain_openai import ChatOpenAI

        return ChatOpenAI(
            model=config.model,
            temperature=config.temperature,
            openai_api_key=config.api_key,
            openai_api_base=config.base_url,
        )

    else:
        raise ValueError("Unsupported LLM configuration")


def get_openai_models():
    """Get the list of OpenAI models."""
    sorted_model_ids = []
    if OPENAI_KEY != "None":
        openai = OpenAI()
        models_list = sorted(
            [model for model in openai.models.list() if model.created >= 1706037777],
            key=lambda x: x.created,
            reverse=True,
        )
        sorted_model_ids = [model.id for model in models_list]
    return sorted_model_ids


def get_anthropic_models():
    """Get the list of Anthropic models."""
    return [
        "claude-3-5-sonnet-20240620",
        "claude-3-opus-20240229",
        "claude-3-sonnet-20240229",
        "claude-3-haiku-20240307",
    ]


def get_ollama_models():
    """Get the list of Ollama models."""
    return ["llama3.1"]


def get_lbl_cborg_models():
    """Get the list of LBNL-hosted models via CBORG."""
    return [
        "lbl/cborg-chat:latest",  # LBL-hosted model
        "lbl/cborg-chat-nano:latest",  # LBL-hosted model
        "lbl/cborg-coder:latest",  # LBL-hosted model
        "openai/chatgpt:latest",  # LBL cloud-hosted model
        "anthropic/claude:latest",  # LBL cloud-hosted model
        "google/gemini:latest",  # LBL cloud-hosted model
    ]


# * App related utilities. **************************************************************************
def generate_random_color():
    """Generate a random color."""
    while True:
        color = "#{:06x}".format(random.randint(0, 0xFFFFFF))
        if color not in PREFIX_COLOR_MAP.values():
            return color


def assign_color_to_prefix(curie):
    """Assign a color to a prefix."""
    prefix = curie.split(":")[0]
    if prefix not in PREFIX_COLOR_MAP:
        PREFIX_COLOR_MAP[prefix] = generate_random_color()
    # print(f"PREFIX_COLOR_MAP={PREFIX_COLOR_MAP}") # To debug
    return PREFIX_COLOR_MAP[prefix]


def visualize_kg(nodes, edges, app: bool = False, output_dir: str = None):
    """Visualize the knowledge graph using pyvis."""
    # Create a PyVis network
    net = Network(
        notebook=False,
        cdn_resources="in_line",
        neighborhood_highlight=True,
        select_menu=True,
        filter_menu=True,
    )

    if not nodes:
        print("No nodes to display.")
        return

    # Track added nodes to avoid duplicates
    added_nodes = set()

    # Add nodes to the network
    for node in nodes:
        # based on prefix (node["id"].split(":")[0]) assign nodes to color

        net.add_node(
            node["id"],
            label=node["label"],
            category=node.get("category", ""),
            title=node.get("id", ""),
            color=assign_color_to_prefix(node["id"]),
        )
        added_nodes.add(node["id"])

    # Add edges to the network
    for edge in edges:
        subject = edge["subject"]["id"]
        object = edge["object"]["id"]

        # Ensure both subject and object nodes are added
        if subject not in added_nodes:
            net.add_node(
                subject,
                label=edge["subject"].get("label", ""),
                category=edge["subject"].get("category", ""),
                title=subject,
                color=assign_color_to_prefix(subject),
            )
            added_nodes.add(subject)
        if object not in added_nodes:
            net.add_node(
                object,
                label=edge["object"].get("label", ""),
                category=edge["object"].get("category", ""),
                title=object,
                color=assign_color_to_prefix(object),
            )
            added_nodes.add(object)

        # Add edge with title (predicate)
        net.add_edge(subject, object, title=edge.get("predicate"))

    # Generate and display the network
    if output_dir:
        output_dir_path = Path(output_dir)
        output_dir_path.mkdir(parents=True, exist_ok=True)
    else:
        output_dir_path = Path.cwd()
    html_file = str(output_dir_path / "knowledge_graph.html")
    net.show(html_file, notebook=False)

    # Open the generated HTML file in the default web browser
    webbrowser.open(html_file)
    if app:
        # Generate the HTML representation
        return net.generate_html()


def split_documents(path):
    """Get the local documents."""
    encodings = ["utf-8", "latin-1", "iso-8859-1"]
    for encoding in encodings:
        try:
            with open(path, "r", encoding=encoding) as doc_file:
                print(f"Reading from file: {path} with encoding: {encoding}")
                doc_object = (Document(page_content=doc_file.read()),)
                splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(chunk_size=300, chunk_overlap=50)
                splits = splitter.split_documents(doc_object)
                return splits
        except UnicodeDecodeError:
            print(f"Failed to decode {path} with encoding: {encoding}")
            continue

    print(f"All decoding attempts failed for file: {path}")
    return None


def get_exisiting_vectorstore():
    """Get the existing vectorstore."""
    if VECTOR_DB_PATH.exists():
        logger.info("Vector database path exists. Loading vectorstore from Chroma.")
        if os.getenv("OPENAI_API_KEY"):
            vectorstore = Chroma(
                embedding_function=OpenAIEmbeddings(show_progress_bar=True), persist_directory=str(VECTOR_STORE)
            )
        else:
            logger.error("OpenAI API key not found. Please set the OPENAI_API_KEY environment variable.")
            vectorstore = None
    else:
        logger.info("Vector database path does not exist. Loading ontology documents.")
        vectorstore = None
    return vectorstore


def create_vectorstore(doc_dir_or_file: Union[Path, str]) -> Chroma:
    """Create a vectorstore from the documents in the doc_dir."""
    if get_exisiting_vectorstore():
        return get_exisiting_vectorstore()
    else:
        logger.info("Vector database path does not exist. Loading ontology documents.")
        doc_dir_or_file_path = Path(doc_dir_or_file)
        list_of_documents = []

        if not doc_dir_or_file_path.is_dir():
            raise ValueError(f"The provided path {doc_dir_or_file} is not a directory.")

        if doc_dir_or_file_path.is_file():
            list_of_documents += split_documents(doc_dir_or_file_path) if split_documents(doc_dir_or_file_path) else []
        elif doc_dir_or_file_path.is_dir():
            for doc_path in doc_dir_or_file_path.rglob("*"):
                if doc_path.is_file() and not doc_path.suffix == ".db":
                    # Process each file here
                    list_of_documents += split_documents(doc_path) if split_documents(doc_path) else []
            vectorstore = Chroma.from_documents(
                documents=list_of_documents,
                embedding=OpenAIEmbeddings(show_progress_bar=True),
                persist_directory=str(VECTOR_STORE),
            )
        else:
            raise ValueError(f"The provided path {doc_dir_or_file} is not a valid file or directory.")
        logger.info("Vectorstore created from documents.")
    return vectorstore



"""Implementation of the DatabaseInterface for DuckDB."""

import logging
import tempfile
import time
from pathlib import Path
from pprint import pprint
from typing import Union

import duckdb
from langchain.agents.agent import AgentExecutor, AgentType
from langchain.tools.retriever import create_retriever_tool
from langchain_community.agent_toolkits import SQLDatabaseToolkit, create_sql_agent
from langchain_community.utilities.sql_database import SQLDatabase
from langchain_ollama import ChatOllama
from sqlalchemy import Engine, create_engine

from kg_chat.config.llm_config import LLMConfig
from kg_chat.constants import VECTOR_DB_PATH
from kg_chat.interface.database_interface import DatabaseInterface
from kg_chat.utils import (
    create_vectorstore,
    get_exisiting_vectorstore,
    llm_factory,
    structure_query,
)

logger = logging.getLogger(__name__)


class DuckDBImplementation(DatabaseInterface):
    """Implementation of the DatabaseInterface for DuckDB."""

    def __init__(self, data_dir: Union[Path, str], llm_config: LLMConfig, doc_dir_or_file: Union[Path, str] = None):
        """Initialize the DuckDB database and the Langchain components."""
        if not data_dir:
            raise ValueError("Data directory is required. This typically contains the KGX tsv files.")
        self.safe_mode = True
        self.data_dir = Path(data_dir)
        self.database_path = self.data_dir / "database/kg_chat.db"
        if not self.database_path.exists():
            self.database_path.parent.mkdir(parents=True, exist_ok=True)
        self.conn: duckdb.DuckDBPyConnection = duckdb.connect(database=str(self.database_path))
        self.llm = llm_factory(llm_config)
        agent_type = AgentType.ZERO_SHOT_REACT_DESCRIPTION
        self.engine: Engine = create_engine(f"duckdb:///{self.database_path}")
        self.db = SQLDatabase(self.engine, view_support=True)
        self.toolkit = SQLDatabaseToolkit(db=self.db, llm=self.llm)
        self.tools = self.toolkit.get_tools()
        if VECTOR_DB_PATH.exists() and get_exisiting_vectorstore():
            vectorstore = get_exisiting_vectorstore()
            retriever = vectorstore.as_retriever(search_kwargs={"k": 1})
            rag_tool = create_retriever_tool(retriever, "VectorStoreRetriever", "Vector Store Retriever")

            self.tools.append(rag_tool)
        elif doc_dir_or_file:
            vectorstore = create_vectorstore(doc_dir_or_file=doc_dir_or_file)
            retriever = vectorstore.as_retriever(search_kwargs={"k": 1})
            rag_tool = create_retriever_tool(retriever, "VectorStoreRetriever", "Vector Store Retriever")

            self.tools.append(rag_tool)
        else:
            logger.info("No vectorstore found or documents provided. Skipping RAG tool creation.")

        self.tool_names = [tool.name for tool in self.tools]

        self.agent: AgentExecutor = create_sql_agent(
            llm=self.llm,
            verbose=True,
            toolkit=self.toolkit,
            agent_type=agent_type,
            agent_executor_kwargs=dict(
                return_intermediate_steps=True,
                handle_parsing_errors=True,
            ),
        )

    def toggle_safe_mode(self, enabled: bool):
        """Toggle safe mode on or off."""
        self.safe_mode = enabled

    def is_safe_command(self, command: str) -> bool:
        """Check if a command is safe to execute."""
        unsafe_keywords = ["DROP", "DELETE", "TRUNCATE", "ALTER", "UPDATE", "INSERT"]
        return not any(keyword in command.upper() for keyword in unsafe_keywords)

    def execute_unsafe_operation(self, operation: callable):
        """Execute an unsafe operation."""
        original_safe_mode = self.safe_mode
        self.safe_mode = False
        try:
            result = operation()
        finally:
            self.safe_mode = original_safe_mode
        return result

    def execute_query(self, query: str):
        """Execute a SQL query against the DuckDB database."""
        if self.safe_mode and not self.is_safe_command(query):
            raise ValueError(f"Unsafe command detected: {query}")
        return self.conn.execute(query).fetchall()

    def clear_database(self):
        """Clear the database."""

        def _clear():
            self.conn.execute("DROP TABLE IF EXISTS edges")
            self.conn.execute("DROP TABLE IF EXISTS nodes")

        return self.execute_unsafe_operation(_clear)

    def get_human_response(self, prompt: str):
        """Get a human response from the database."""
        response = self.agent.invoke(
            {
                "input": prompt,
                "tools": self.tools,
                "tool_names": self.tool_names,
            }
        )
        return response["output"]

    def get_structured_response(self, prompt: str):
        """Get a structured response from the database."""
        if isinstance(self.llm, ChatOllama):
            if "show me" in prompt.lower():
                self.llm.format = "json"

        #     tool_names = [tool.name for tool in self.toolkit.get_tools()] + ["kg_retriever"]

        #     structured_query = get_sql_agent_prompt_template().format(
        #         input=prompt,
        #         tools=self.tools,
        #         tool_names=tool_names,
        #         agent_scratchpad=None,
        #     )
        # else:
        #     structured_query = {"input": structure_query(prompt)}

        # response = self.agent.invoke(structured_query)
        response = self.agent.invoke(
            {
                "input": structure_query(prompt),
                "tools": self.tools,
                "tool_names": self.tool_names,
            }
        )
        return response["output"]

    def create_edges(self, edges):
        """Create edges in the database."""

        def _create_edges():
            self.conn.executemany(
                "INSERT INTO edges (subject, predicate, object) VALUES (?, ?, ?)",
                edges,
            )

        return self.execute_unsafe_operation(_create_edges)

    def create_nodes(self, nodes):
        """Create nodes in the database."""

        def _create_nodes():
            self.conn.executemany(
                "INSERT INTO nodes (id, category, label) VALUES (?, ?, ?)",
                nodes,
            )

        return self.execute_unsafe_operation(_create_nodes)

    def show_schema(self):
        """Show the schema of the database."""
        result = self.conn.execute("PRAGMA show_tables").fetchall()
        return pprint(result)

    def execute_query_using_langchain(self, prompt: str):
        """Execute a query against the database using Langchain."""
        # response = self.agent.invoke(prompt)
        response = self.agent.invoke(
            {
                "input": prompt,
                "tools": self.tools,
                "tool_names": self.tool_names,
            }
        )
        return response["output"]

    def load_kg(self):
        """Load the Knowledge Graph into the database."""

        def _load_kg():
            # Clear the existing database
            print("Clearing the existing database...")
            self.clear_database()
            print("Database cleared.")

            # Create tables
            self.conn.execute(
                """
                CREATE TABLE nodes (
                    id VARCHAR PRIMARY KEY,
                    category VARCHAR,
                    label VARCHAR
                )
            """
            )
            self.conn.execute(
                """
                CREATE TABLE edges (
                    subject VARCHAR,
                    predicate VARCHAR,
                    object VARCHAR
                )
            """
            )

            # Import nodes
            print("Starting to import nodes...")
            start_time = time.time()
            self._import_nodes()

            elapsed_time_seconds = time.time() - start_time

            if elapsed_time_seconds >= 3600:
                elapsed_time_hours = elapsed_time_seconds / 3600
                print(f"Nodes import completed in {elapsed_time_hours:.2f} hours.")
            else:
                elapsed_time_minutes = elapsed_time_seconds / 60
                print(f"Nodes import completed in {elapsed_time_minutes:.2f} minutes.")

            # Import edges
            print("Starting to import edges...")
            start_time = time.time()
            self._import_edges()

            elapsed_time_seconds = time.time() - start_time
            if elapsed_time_seconds >= 3600:
                elapsed_time_hours = elapsed_time_seconds / 3600
                print(f"Edges import completed in {elapsed_time_hours:.2f} hours.")
            else:
                elapsed_time_minutes = elapsed_time_seconds / 60
                print(f"Edges import completed in {elapsed_time_minutes:.2f} minutes.")

            # Create indexes for faster querying
            print("Creating indexes...")
            start_time = time.time()
            self.conn.execute("CREATE INDEX idx_nodes_id ON nodes(id);")
            # self.conn.execute("CREATE INDEX idx_nodes_category ON nodes(category);")
            self.conn.execute("CREATE INDEX idx_edges_subject ON edges(subject);")
            # self.conn.execute("CREATE INDEX idx_edges_predicate ON edges(predicate);")
            self.conn.execute("CREATE INDEX idx_edges_object ON edges(object);")
            elapsed_time_seconds = time.time() - start_time
            if elapsed_time_seconds >= 3600:
                elapsed_time_hours = elapsed_time_seconds / 3600
                print(f"Index creation completed in {elapsed_time_hours:.2f} hours.")
            else:
                elapsed_time_minutes = elapsed_time_seconds / 60
                print(f"Index creation completed completed in {elapsed_time_minutes:.2f} minutes.")

            print("Import process finished.")

        return self.execute_unsafe_operation(_load_kg)

    def _import_nodes(self):
        columns_of_interest = ["id", "category", "name", "description"]
        nodes_filepath = Path(self.data_dir) / "nodes.tsv"

        with open(nodes_filepath, "r") as nodes_file:
            header_line = nodes_file.readline().strip().split("\t")
            column_indexes = {col: idx for idx, col in enumerate(header_line) if col in columns_of_interest}

            # Determine which label column to use
            label_column = "name" if "name" in column_indexes else "description"

            with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_nodes_file:
                # Write the header line with the correct label column
                temp_nodes_file.write("\t".join(["id", "category", label_column]) + "\n")
                for line in nodes_file:
                    columns = line.strip().split("\t")
                    if len(columns) > max(column_indexes.values()):
                        node_id = columns[column_indexes["id"]]
                        node_category = columns[column_indexes["category"]]
                        node_label = columns[column_indexes[label_column]]
                        temp_nodes_file.write(f"{node_id}\t{node_category}\t{node_label}\n")
                temp_nodes_file.flush()

                # Load data from temporary file into DuckDB
                self.conn.execute(f"COPY nodes FROM '{temp_nodes_file.name}' (DELIMITER '\t', HEADER)")

    def _import_edges(self):
        edge_column_of_interest = ["subject", "predicate", "object"]
        edges_filepath = Path(self.data_dir) / "edges.tsv"
        with open(edges_filepath, "r") as edges_file:
            header_line = edges_file.readline().strip().split("\t")
            column_indexes = {col: idx for idx, col in enumerate(header_line) if col in edge_column_of_interest}
            with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_edges_file:
                temp_edges_file.write("\t".join(edge_column_of_interest) + "\n")
                for line in edges_file:
                    columns = line.strip().split("\t")
                    if len(columns) == (max(column_indexes.values()) + 1):
                        subject = columns[column_indexes["subject"]]
                        predicate = columns[column_indexes["predicate"]]
                        object = columns[column_indexes["object"]]
                        temp_edges_file.write(f"{subject}\t{predicate}\t{object}\n")
                temp_edges_file.flush()

                # Load data from temporary file into DuckDB
                self.conn.execute(f"COPY edges FROM '{temp_edges_file.name}' (DELIMITER '\t', HEADER)")



"""Configuration for the LLM model."""

from pydantic import BaseModel


class LLMConfig(BaseModel):
    """Configuration for the LLM model."""

    model: str
    api_key: str = None
    temperature: float = 0.0


class OpenAIConfig(LLMConfig):
    """Configuration for OpenAI LLM model."""

    def __init__(self, **data):
        """Initialize the OpenAI LLM configuration."""
        super().__init__(**data)
        if self.model.startswith("o1"):
            self.temperature = 1.0


class OllamaConfig(LLMConfig):
    """Configuration for Ollama LLM model."""

    format: str = None
    pass


class AnthropicConfig(LLMConfig):
    """Configuration for Anthropic LLM model."""

    pass


class CBORGConfig(LLMConfig):
    """Configuration for CBORG LLM model."""

    base_url: str = "https://api.cborg.lbl.gov"  # Local clients can also use https://api-local.cborg.lbl.gov
